# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import
import os
from astrometry.util.fits import fits_table
import numpy as np
import logging
import tempfile

import sys
py3 = (sys.version_info[0] >= 3)
if py3:
    from urllib.parse import urljoin
else:
    from urlparse import urljoin

fitsio = None
try:
    import fitsio
except:
    try:
        import pyfits
    except ImportError:
        try:
            from astropy.io import fits as pyfits
        except ImportError:
            raise ImportError("Cannot import either pyfits or astropy.io.fits")

from .common import *
from .dr7 import *
from .yanny import *
from astrometry.util.run_command import run_command

class Frame(SdssFile):
    def __init__(self, *args, **kwargs):
        super(Frame, self).__init__(*args, **kwargs)
        self.filetype = 'frame'
        self.image = None

        self.image_proxy = None

    def getImageShape(self):
        if self.image_proxy is not None:
            # fitsio fits.FITSHDU object
            H,W = self.image_proxy.get_info()['dims']
            H = int(H)
            W = int(W)
        else:
            H,W = self.image.shape
        return H,W

    def getImageSlice(self, slice):
        if self.image_proxy is not None:
            #print 'reading slice from image proxy:', slice
            return self.image_proxy[slice]
        return self.image[slice]

    #def __str__(self):
    def getImage(self):
        if self.image is None and self.image_proxy is not None:
            self.image = self.image_proxy.read()
            self.image_proxy = None
        return self.image
    def getHeader(self):
        return self.header
    def getAsTrans(self):
        return self.astrans
    def getCalibVec(self):
        return self.calib

    def getSkyAt(self, x, y):
        skyim = self.sky
        (sh,sw) = skyim.shape
        if sw != 256:
            skyim = skyim.T
        (sh,sw) = skyim.shape
        xi = np.round(self.skyxi[x]).astype(int)
        yi = np.round(self.skyyi[y]).astype(int)
        yi = np.minimum(yi,sh-1)
        return skyim[yi,xi]
    
    def getSky(self):
        skyim = self.sky
        (sh,sw) = skyim.shape
        if sw != 256:
            skyim = skyim.T
        (sh,sw) = skyim.shape
        xi = np.round(self.skyxi).astype(int)
        yi = np.round(self.skyyi).astype(int)
        yi = np.minimum(yi,sh-1)
        assert(all(xi >= 0) and all(xi < sw))
        assert(all(yi >= 0) and all(yi < sh))
        XI,YI = np.meshgrid(xi, yi)
        # Nearest-neighbour interpolation -- we just need this
        # for approximate invvar.
        bigsky = skyim[YI,XI]
        return bigsky

    def getInvvar(self, psfield, bandnum, ignoreSourceFlux=False,
                  sourceFlux=None, constantSkyAt=None):
        '''
        If constantSkyAt = (x,y) (INTEGERS!),
        returns a scalar (rather than a np.array) of the invvar at that point.
        
        NOTE that this does NOT blank out masked pixels; use, eg,

        fpM = sdss.readFpM(run, camcol, field, bandname)
        for plane in [ 'INTERP', 'SATUR', 'CR', 'GHOST' ]:
            fpM.setMaskedPixels(plane, invvar, 0, roi=roi)
        '''
        calibvec = self.getCalibVec()

        if constantSkyAt:
            x,y = constantSkyAt
            calibvec = calibvec[x]
            sky = self.getSkyAt(x,y)
            if ignoreSourceFlux:
                dn = sky
            elif sourceFlux is None:
                image = self.getImage()
                dn = (image[y,x] / calibvec) + sky
            else:
                dn = (sourceFlux / calibvec) + sky
        else:
            bigsky = self.getSky()
            if ignoreSourceFlux:
                dn = bigsky
            elif sourceFlux is None:
                image = self.getImage()
                dn = (image / calibvec) + bigsky
            else:
                dn = (sourceFlux / calibvec) + bigsky

        gain = psfield.getGain(bandnum)
        # Note, "darkvar" includes dark current *and* read noise.
        darkvar = psfield.getDarkVariance(bandnum)
        dnvar = (dn / gain) + darkvar
        invvar = 1./(dnvar * calibvec**2)
        return invvar

class PhotoObj(SdssFile):
    def __init__(self, *args, **kwargs):
        super(PhotoObj, self).__init__(*args, **kwargs)
        self.filetype = 'photoObj'
        self.table = None
    def getTable(self):
        return self.table

class runlist(object):
    pass

class DR8(DR7):
    _lup_to_mag_b = np.array([1.4e-10, 0.9e-10, 1.2e-10, 1.8e-10, 7.4e-10])
    _two_lup_to_mag_b = 2.*_lup_to_mag_b
    _ln_lup_to_mag_b = np.log(_lup_to_mag_b)

    '''
    From
    http://data.sdss3.org/datamodel/glossary.html#asinh

    m = -(2.5/ln(10))*[asinh(f/2b)+ln(b)].

    The parameter b is a softening parameter measured in maggies, and
    for the [u, g, r, i, z] bands has the values
    [1.4, 0.9, 1.2, 1.8, 7.4] x 1e-10
    '''
    @staticmethod
    def luptitude_to_mag(Lmag, bandnum, badmag=25):
        if bandnum is None:
            # assume Lmag is broadcastable to a 5-vector
            twobi = DR8._two_lup_to_mag_b
            lnbi = DR8._ln_lup_to_mag_b
        else:
            twobi = DR8._two_lup_to_mag_b[bandnum]
            lnbi = DR8._ln_lup_to_mag_b[bandnum]
        # MAGIC -1.08.... = -2.5/np.log(10.)
        f = np.sinh(Lmag/-1.0857362047581294 - lnbi) * twobi
        # prevent log10(-flux)
        mag = np.zeros_like(f) + badmag
        I = (f > 0)
        mag[I] = -2.5 * np.log10(f[I])
        return mag

    @staticmethod
    def nmgy_to_mag(nmgy):
        return 22.5 - 2.5 * np.log10(nmgy)

    def getDRNumber(self):
        return 8

    def useLocalTree(self, photoObjs=None, resolve=None):
        if photoObjs is None:
            photoObjs = os.environ['BOSS_PHOTOOBJ']
        redux = os.environ['PHOTO_REDUX']
        if resolve is None:
            resolve = os.environ['PHOTO_RESOLVE']
        
        self.filenames.update(
            photoObj = os.path.join(photoObjs, '%(rerun)s', '%(run)i', '%(camcol)i',
                                    'photoObj-%(run)06i-%(camcol)i-%(field)04i.fits'),
            frame = os.path.join(photoObjs, 'frames', '%(rerun)s', '%(run)i', '%(camcol)i',
                                    'frame-%(band)s-%(run)06i-%(camcol)i-%(field)04i.fits.bz2'),
            photoField = os.path.join(photoObjs, '%(rerun)s', '%(run)i',
                                      'photoField-%(run)06i-%(camcol)i.fits'),
            psField = os.path.join(redux, '%(rerun)s', '%(run)i', 'objcs', '%(camcol)i',
                                   'psField-%(run)06i-%(camcol)i-%(field)04i.fit'),
            fpM = os.path.join(redux, '%(rerun)s', '%(run)i', 'objcs', '%(camcol)i',
                               'fpM-%(run)06i-%(band)s%(camcol)i-%(field)04i.fit.gz'),
            window_flist = os.path.join(resolve, 'window_flist.fits'),
            )
        # use fpM files compressed
        try:
            del self.dassuffix['fpM']
        except:
            pass
        try:
            del self.processcmds['fpM']
        except:
            pass

    def saveUnzippedFiles(self, basedir):
        self.unzip_dir = basedir

    def setFitsioReadBZ2(self, to=True):
        '''
        Call this if fitsio supports reading .bz2 files directly.
        '''
        self.readBz2 = to
    
    def __init__(self, **kwargs):
        '''
        Useful kwargs:
        
        basedir : (string) - local directory where data will be stored.
        '''
        DR7.__init__(self, **kwargs)

        self.unzip_dir = None
        self.readBz2 = False

        # Local filenames
        self.filenames.update({
            'frame': 'frame-%(band)s-%(run)06i-%(camcol)i-%(field)04i.fits.bz2',
            'idR': 'idR-%(run)06i-%(band)s-%(camcol)i-%(field)04i.fits',
            'photoObj': 'photoObj-%(run)06i-%(camcol)i-%(field)04i.fits',
            'photoField': 'photoField-%(run)06i-%(camcol)i.fits',
            'window_flist': 'window_flist.fits',
            })

        # URLs on DAS server
        self.dasurl = 'http://data.sdss3.org/sas/dr8/groups/boss/'
        self.daspaths = {
            'idR': 'photo/data/%(run)i/fields/%(camcol)i/idR-%(run)06i-%(band)s%(camcol)i-%(field)04i.fit.Z',
            'fpObjc': 'photo/redux/%(rerun)s/%(run)i/objcs/%(camcol)i/fpObjc-%(run)06i-%(camcol)i-%(field)04i.fit',
            # DR8 frames are no longer available on DAS.
            'frame': '/sas/dr9/boss/photoObj/frames/%(rerun)s/%(run)i/%(camcol)i/frame-%(band)s-%(run)06i-%(camcol)i-%(field)04i.fits.bz2',
            #'frame': 'photoObj/frames/%(rerun)s/%(run)i/%(camcol)i/frame-%(band)s-%(run)06i-%(camcol)i-%(field)04i.fits.bz2',
            'photoObj': 'photoObj/%(rerun)s/%(run)i/%(camcol)i/photoObj-%(run)06i-%(camcol)i-%(field)04i.fits',
            'psField': 'photo/redux/%(rerun)s/%(run)i/objcs/%(camcol)i/psField-%(run)06i-%(camcol)i-%(field)04i.fit',
            'photoField': 'photoObj/%(rerun)s/%(run)i/photoField-%(run)06i-%(camcol)i.fits',
            'fpM': 'photo/redux/%(rerun)s/%(run)i/objcs/%(camcol)i/fpM-%(run)06i-%(band)s%(camcol)i-%(field)04i.fit.gz',
            'fpAtlas': 'photo/redux/%(rerun)s/%(run)i/objcs/%(camcol)i/fpAtlas-%(run)06i-%(camcol)i-%(field)04i.fit',
            'window_flist': 'resolve/2010-05-23/window_flist.fits',
            }

        self.dassuffix = {
            #'frame': '.bz2',
            'fpM': '.gz',
            'idR': '.Z',
            }

        # called in retrieve()
        self.processcmds = {
            'fpM': 'gunzip -cd %(input)s > %(output)s',
            'idR': 'gunzip -cd %(input)s > %(output)s',
            }

        self.postprocesscmds = {
            'frame': 'TMPFILE=$(mktemp %(output)s.tmp.XXXXXX) && bunzip2 -cd %(input)s > $TMPFILE && mv $TMPFILE %(output)s',
            }
        
        y = read_yanny(self._get_runlist_filename())
        y = y['RUNDATA']
        rl = runlist()
        rl.run = np.array(y['run'])
        rl.startfield = np.array(y['startfield'])
        rl.endfield = np.array(y['endfield'])
        rl.rerun = np.array(y['rerun'])
        #print 'Rerun type:', type(rl.rerun), rl.rerun.dtype
        self.runlist = rl

        self.logger = logging.getLogger('astrometry.sdss.DR%i' %
                                        self.getDRNumber())
        #self.logger.debug('debug test')
        #self.logger.info('info test')
        #self.logger.warning('warning test')

    def _unzip_frame(self, fn, run, camcol):
        if self.readBz2:
            return None,True

        # No, PJM reported that pyfits failed on SDSS frame*.bz2 files
        # if not fitsio:
        #    # pyfits can read .bz2
        #    return None,True

        tempfn = None
        keep = False

        filetype = 'frame'
        if not(filetype in self.postprocesscmds and fn.endswith('.bz2')):
            return None,True
        
        cmd = self.postprocesscmds[filetype]

        if self.unzip_dir is not None:
            udir = os.path.join(self.unzip_dir, '%i' % run, '%i' % camcol)
            if not os.path.exists(udir):
                try:
                    os.makedirs(udir)
                except:
                    pass
            tempfn = os.path.join(udir, os.path.basename(fn).replace('.bz2', ''))
            #print 'Checking', tempfn
            if os.path.exists(tempfn):
                print('File exists:', tempfn)
                return tempfn,True
            else:
                print('Saving to', tempfn)
                keep = True

        else:
            fid,tempfn = tempfile.mkstemp()
            os.close(fid)

        cmd = cmd % dict(input = fn, output = tempfn)
        self.logger.debug('cmd: %s' % cmd)
        print('command:', cmd)
        (rtn,out,err) = run_command(cmd)
        if rtn:
            print('Command failed: command', cmd)
            print('Output:', out)
            print('Error:', err)
            print('Return val:', rtn)
            raise RuntimeError('Command failed (return val %i): %s' % (rtn, cmd))

        print(out)
        print(err)
        return tempfn,keep


        
        
    def _get_runlist_filename(self):
        return self._get_data_file('runList-dr8.par')

    # read a data file describing the DR8 data
    def _get_data_file(self, fn):
        return os.path.join(os.path.dirname(__file__), fn)

    def get_rerun(self, run, field=None):
        I = (self.runlist.run == run)
        if field is not None:
            I *= (self.runlist.startfield <= field) * (self.runlist.endfield >= field)
        I = np.flatnonzero(I)
        reruns = np.unique(self.runlist.rerun[I])
        #print 'Run', run, '-> reruns:', reruns
        if len(reruns) == 0:
            return None
        return reruns[-1]

    def get_url(self, filetype, run, camcol, field, band=None, rerun=None):
        if rerun is None:
            rerun = self.get_rerun(run, field)
        path = self.daspaths[filetype]
        url = urljoin(self.dasurl, path % dict(
            run=run, camcol=camcol, field=field, rerun=rerun, band=band))
        return url
    
    def retrieve(self, filetype, run, camcol, field=None, band=None, skipExisting=True,
                 tempsuffix='.tmp', rerun=None):

        outfn = self.getPath(filetype, run, camcol, field, band,
                             rerun=rerun)
        print('Checking for file', outfn)
        if outfn is None:
            return None
        if skipExisting and os.path.exists(outfn):
            #print('Exists')
            return outfn
        outdir = os.path.dirname(outfn)
        if not os.path.exists(outdir):
            try:
                os.makedirs(outdir)
            except:
                pass
        url = self.get_url(filetype, run, camcol, field, band=band, rerun=rerun)
        #print 'Did not find file:', outfn
        print('Retrieving from URL:', url)
        if self.curl:
            cmd = "curl -o '%(outfn)s' '%(url)s'"
        else:
            cmd = "wget --continue -nv -O %(outfn)s '%(url)s'"

        # suffix to add to the downloaded filename
        suff = self.dassuffix.get(filetype, '')

        oo = outfn + suff
        if tempsuffix is not None:
            oo += tempsuffix
        
        cmd = cmd % dict(outfn=oo, url=url)
        self.logger.debug('cmd: %s' % cmd)
        (rtn,out,err) = run_command(cmd)
        if rtn:
            print('Command failed: command', cmd)
            print('Output:', out)
            print('Error:', err)
            print('Return val:', rtn)
            return None

        if tempsuffix is not None:
            #
            self.logger.debug('Renaming %s to %s' % (oo, outfn+suff))
            os.rename(oo, outfn + suff)

        if filetype in self.processcmds:
            cmd = self.processcmds[filetype]
            cmd = cmd % dict(input = outfn + suff, output = outfn)
            self.logger.debug('cmd: %s' % cmd)
            (rtn,out,err) = run_command(cmd)
            if rtn:
                print('Command failed: command', cmd)
                print('Output:', out)
                print('Error:', err)
                print('Return val:', rtn)
                return None

        return outfn

    def readPhotoObj(self, run, camcol, field, filename=None):
        obj = PhotoObj(run, camcol, field)
        if filename is None:
            fn = self.getPath('photoObj', run, camcol, field)
        else:
            fn = filename
        obj.table = fits_table(fn)
        return obj

    def readFrame(self, run, camcol, field, band, filename=None):
        '''
        http://data.sdss3.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
        '''
        f = Frame(run, camcol, field, band)
        # ...
        if filename is None:
            fn = self.getPath('frame', run, camcol, field, band)
        else:
            fn = filename

        # optionally bunzip2 the frame file.
        tempfn,keep = self._unzip_frame(fn, run, camcol)
        if tempfn is not None:
            fn = tempfn

        if fitsio:

            print('Frame filename', fn)
            # eg /clusterfs/riemann/raid006/dr10/boss/photoObj/frames/301/2825/1/frame-u-002825-1-0126.fits.bz2

            F = fitsio.FITS(fn, lower=True)
            f.header = F[0].read_header()
            # Allow later reading of just the pixels of interest.
            f.image_proxy = F[0]
            f.calib = F[1].read()
            sky = F[2].read_columns(['allsky', 'xinterp', 'yinterp'])
            #print 'sky', type(sky)
            # ... supposed to be a recarray, but it's not...
            f.sky, f.skyxi, f.skyyi = sky.tolist()[0]
            tab = fits_table(F[3].read())
            if not keep and tempfn is not None:
                os.remove(tempfn)

        else:
            p = pyfits.open(fn)
            # in nanomaggies
            f.image = p[0].data
            f.header = p[0].header
            # converts counts -> nanomaggies
            f.calib = p[1].data
            # table with val,x,y -- binned; use bilinear interpolation to expand
            sky = p[2].data
            # table -- asTrans structure
            tab = fits_table(p[3].data)

            f.sky = sky.field('allsky')[0]
            f.skyxi = sky.field('xinterp')[0]
            f.skyyi = sky.field('yinterp')[0]

        #print 'sky shape', f.sky.shape
        if len(f.sky.shape) != 2:
            f.sky = f.sky.reshape((-1, 256))
        assert(len(tab) == 1)
        tab = tab[0]
        # DR7 has NODE, INCL in radians...
        f.astrans = AsTrans(run, camcol, field, band,
                            node=np.deg2rad(tab.node), incl=np.deg2rad(tab.incl),
                            astrans=tab, cut_to_band=False)
                            
        return f
    



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib
matplotlib.use('Agg')

from astrometry.util.index import *
from astrometry.util.plotutils import *
from astrometry.libkd.spherematch import match
from astrometry.util.starutil_numpy import *
from astrometry.util.fits import *

from optparse import *

from pylab import *
from numpy import *
import os

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option('-p', '--prefix', dest='prefix', help='Prefix for output plot names')
    parser.add_option('-r', '--range', dest='range', type='float', help='Set search radius range (in arcsec) of stars-1 plot, default 15')
    parser.set_defaults(prefix='', range=15.)
    opt,args = parser.parse_args()

    if 'plots' in args: # DEBUG!
        cat = fits_table('cat2.fits')
        xyz = radectoxyz(cat.ra, cat.dec)
        R = 15.
        inds,dists = match(xyz, xyz, deg2rad(R/3600.))
        notself = (inds[:,0] != inds[:,1])
        clf()
        hist(rad2deg(dists[notself]) * 3600., 200)
        title('ImSim reference catalog')
        xlabel('Distance between pairs of sources (arcsec)')
        ylabel('Counts')
        xlim(0, R)
        savefig('cat-stars-1.png')

        cat = fits_table('stars3.fits')
        xyz = radectoxyz(cat.ra, cat.dec)
        R = 15.
        inds,dists = match(xyz, xyz, deg2rad(R/3600.))
        notself = (inds[:,0] != inds[:,1])
        clf()
        hist(rad2deg(dists[notself]) * 3600., 200)
        title('ImSim reference catalog -- stars only')
        xlabel('Distance between pairs of sources (arcsec)')
        ylabel('Counts')
        xlim(0, R)
        savefig('cat-stars-2.png')

        I1 = inds[notself,0]
        I2 = inds[notself,1]
        clf()
        RA = cat.ra + (cat.ra > 180)*-360
        dra = RA[I1]-RA[I2]
        ddec = cat.dec[I1]-cat.dec[I2]
        #plot(dra, ddec, 'r.')
        (H,xe,ye) = histogram2d(dra, ddec, bins=(200,200))
        H=H.T
        imshow(H, extent=(min(xe), max(xe), min(ye), max(ye)), aspect='auto',
               interpolation='nearest', origin='lower', cmap=antigray)

        xlabel('dRA (deg)')
        ylabel('dDec (deg)')
        axis('equal')
        savefig('cat-stars-4.png')

        cat = fits_table('gals2.fits')
        xyz = radectoxyz(cat.ra, cat.dec)
        R = 15.
        inds,dists = match(xyz, xyz, deg2rad(R/3600.))
        notself = (inds[:,0] != inds[:,1])
        clf()
        hist(rad2deg(dists[notself]) * 3600., 200)
        title('ImSim reference catalog -- galaxies only')
        xlabel('Distance between pairs of sources (arcsec)')
        ylabel('Counts')
        xlim(0, R)
        savefig('cat-stars-3.png')
        sys.exit(0)

    for indfn in args:
        print('Reading index', indfn)
        null = None
        I = index_load(indfn, 0, null)
        print('Loaded.')
        NS = index_nstars(I)
        NQ = index_nquads(I)
        print('Index has', NS, 'stars and', NQ, 'quads')
        DQ = index_get_quad_dim(I)
        print('Index has "quads" with %i stars' % (DQ))
        DC = index_get_quad_dim(I)
        print('Index has %i-dimensional codes' % (DC))

        iname = os.path.basename(I.indexname).replace('.fits', '')
        # stars
        print('Getting stars...')
        stars = index_get_stars(I)
        print(stars.shape)

        ra,dec = xyztoradec(stars)
        ra += (ra > 180)*-360

        # FIXME --!
        #ra *= cos(deg2rad(ra))
        rng = [[-10,10],[-10,10]]
        clf()
        (H,xe,ye) = histogram2d(ra, dec, bins=(100,100), range=rng)
        H=H.T
        binarea = (xe[1]-xe[0])*(ye[1]-ye[0])
        print('Bin area:', binarea, 'deg^2')
        binarea *= 3600.
        print(binarea, 'arcmin^2')
        imshow(H/binarea, extent=(min(xe), max(xe), min(ye), max(ye)),
               aspect='auto',
               interpolation='nearest', origin='lower', cmap=antigray)
        colorbar()
        axis('equal')
        xlabel('RA (deg)')
        ylabel('Dec (deg)')
        title('Reference source density in %s' % iname)
        savefig(opt.prefix + 'stars-2.png')

        R = opt.range
        print('Finding pairs within', R, 'arcsec')
        inds,dists = match(stars, stars, deg2rad(R/3600.))
        print('inds', inds.shape, 'dists', dists.shape)

        notself = (inds[:,0] != inds[:,1])
        clf()
        hist(rad2deg(dists[notself]) * 3600., 200)
        xlabel('Star pair distances (arcsec)')
        ylabel('Counts')
        xlim(0, R)
        savefig(opt.prefix + 'stars-1.png')


        # codes
        print('Getting codes...')
        codes = index_get_codes(I)
        print('shape', codes.shape)

        # code slices
        cx = codes[:,0]
        cy = codes[:,1]
        dx = codes[:,2]
        dy = codes[:,3]
        clf()
        (H,xe,ye) = histogram2d(cx, cy, bins=(100,100))
        H=H.T
        imshow(H, extent=(min(xe), max(xe), min(ye), max(ye)), aspect='auto',
               interpolation='nearest', origin='lower', cmap=antigray)
        axis('equal')
        xlabel('cx')
        ylabel('cy')
        savefig(opt.prefix + 'codes-1.png')

        clf()
        (H,xe,ye) = histogram2d(append(cx, dx), append(cy, dy), bins=(100,100))
        H=H.T
        imshow(H, extent=(min(xe), max(xe), min(ye), max(ye)), aspect='auto',
               interpolation='nearest', origin='lower', cmap=antigray)
        axis('equal')
        xlabel('cx, dx')
        ylabel('cy, dy')
        savefig(opt.prefix + 'codes-2.png')

        clf()
        (H,xe,ye) = histogram2d(cx, dx, bins=(100,100))
        H=H.T
        imshow(H, extent=(min(xe), max(xe), min(ye), max(ye)), aspect='auto',
               interpolation='nearest', origin='lower', cmap=antigray)
        axis('equal')
        xlabel('cx')
        ylabel('dx')
        savefig(opt.prefix + 'codes-3.png')

        clf()
        xx = append(cx, dx)
        yy = append(cy, dy)
        (H,xe,ye) = histogram2d(append(xx, 1.0-xx), append(yy, 1.0-yy), bins=(100,100))
        H=H.T
        imshow(H, extent=(min(xe), max(xe), min(ye), max(ye)), aspect='auto',
               interpolation='nearest', origin='lower', cmap=antigray)
        axis('equal')
        xlabel('cx, dx')
        ylabel('cy, dy')
        title('duplicated for A-B swap')
        savefig(opt.prefix + 'codes-4.png')

        for pnum,arrs in [(5, [cx,cy,dx,dy]),
                          (6, [append(cx, 1-cx), append(cy, 1-cy), append(dx, 1-dx), append(dy, 1-dy)])]:
            clf()
            anames = ['cx', 'cy', 'dx', 'dy']
            sp = 0
            for i,a1 in enumerate(arrs):
                for j,a2 in enumerate(arrs):
                    sp += 1
                    if j > i:
                        continue
                    subplot(4,4, sp)
                    if i == j:
                        hist(a1, 100, range=[-0.3, 1.3])
                    else:
                        (H,xe,ye) = histogram2d(a2, a1, bins=(100,100), range=[[-0.3,1.3],[-0.3,1.3]])
                        H=H.T
                        imshow(H, extent=(min(xe), max(xe), min(ye), max(ye)), aspect='auto',
                               interpolation='nearest', origin='lower', cmap=antigray)
                        axis([-0.3,1.3,-0.3,1.3])
                        axis('scaled')
                    xticks([],[])
                    yticks([],[])
                    if i == 3:
                        xlabel(anames[j])
                    if j == 0:
                        ylabel(anames[i])
            savefig(opt.prefix + 'codes-%i.png' % pnum)


        index_free(I)
        



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

## This file includes code from Python 3.8's multiprocessing module.
#
# Copyright (c) 2006-2008, R Oudkerk
# Licensed to PSF under a Contributor Agreement.
#

'''
This file provides a subclass of the multiprocessing.Pool class that
tracks the CPU and Wall time of worker processes, as well as time and
I/O spent in pickling data.

It also provides an astrometry.util.ttime Measurement class,
TimingPoolMeas, that records & reports the pool's worker CPU time.

    dpool = TimingPool(4, taskqueuesize=4)
    dmup = multiproc.multiproc(pool=dpool)
    Time.add_measurement(TimingPoolMeas(dpool))
'''

#
# Pool has an _inqueue (_quick_put) and _outqueue (_quick_get)
# and _taskqueue:
#
#   pool.map()  ---> sets cache[]
#               ---> put work on taskqueue
#       handle_tasks thread  ---> gets work from taskqueue
#                            ---> puts work onto inqueue
#       worker threads       ---> get work from inqueue
#                            ---> put results into outqueue
#       handle_results thread --> gets results from outqueue
#                             --> sets cache[]
#
#       meanwhile, handle_workers thread creates new workers as needed.
#
# map() etc add themselves to cache[jobid] = self
# and place work on the task queue.
#
# _handle_tasks pulls tasks from the _taskqueue and puts them
#     on the _inqueue.
#
# _handle_results pulls results from the _outqueue (job,i,obj)
#     and calls cache[job].set(i, obj)
#     (cache[job] is an ApplyResult / MapResult, etc.)
#
# worker threads run the worker() function:
#   run initializer
#   while true:
#     pull task from inqueue
#     job,i,func,arg,kwargs = task
#     put (job,i,result) on outqueue
#
# _inqueue,_outqueue are SimpleQueue (queues.py)



# To capture the pickle traffic, we subclass the Connection, Pipe, and Queue
# classes.

# To capture the worker CPU time, we intercept the messages put on the inqueue
# and fetched from the outqueue --
# The inqueue includes job ids, the function to called, and args.  We wrap the
# function in time_func() object, so the worker records the CPU time, and
# returns a tuple of the original result and the timing information.
# When reading from the outqueue, we peel off the timing information.

import os
import sys
import time
import struct
import io

import multiprocessing
from multiprocessing.queues import SimpleQueue
from multiprocessing.connection import Connection
from multiprocessing.pool import Pool
from multiprocessing import context
try:
    ForkingPickler = context.reduction.ForkingPickler
except:
    # python3.5
    ForkingPickler = multiprocessing.reduction.ForkingPickler

from astrometry.util.ttime import CpuMeas

class TimingConnection(Connection):
    '''
    A multiprocessing *Connection* subclass that keeps track of how
    many objects and how many bytes are pickled, and how long that
    takes.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ptime = 0.
        self.uptime = 0.
        self.pbytes = 0
        self.upbytes = 0
        self.pobjs = 0
        self.upobjs = 0

    def stats(self):
        '''
        Returns statistics of the objects handled by this connection.

        This method is new in this subclass.
        '''
        return dict(pickle_objs = self.pobjs,
                    pickle_bytes = self.pbytes,
                    pickle_megabytes = 1e-6 * self.pbytes,
                    pickle_cputime = self.ptime,
                    unpickle_objs = self.upobjs,
                    unpickle_bytes = self.upbytes,
                    unpickle_megabytes = 1e-6 * self.upbytes,
                    unpickle_cputime = self.uptime)

    def recv(self):
        """Receive a (picklable) object.

        This is overriding a core method in the Connection superclass."""
        self._check_closed()
        self._check_readable()
        t0 = time.time()
        buf = self._recv_bytes()
        buf = buf.getbuffer()
        n = len(buf)
        obj = ForkingPickler.loads(buf)
        dt = time.time() - t0
        self.upbytes += n
        self.uptime += dt
        self.upobjs += 1
        return obj

    def send(self, obj):
        """Send a (picklable) object"""
        self._check_closed()
        self._check_writable()
        t0 = time.time()
        bb = ForkingPickler.dumps(obj)
        dt = time.time() - t0
        self.pbytes += len(bb)
        self.ptime += dt
        self.pobjs += 1
        return self._send_bytes(bb)

    # Borrowed from the python3.8 superclass -- handles > 2GB pickles.
    def _send_bytes(self, buf):
        n = len(buf)
        if n > 0x7fffffff:
            pre_header = struct.pack("!i", -1)
            header = struct.pack("!Q", n)
            self._send(pre_header)
            self._send(header)
            self._send(buf)
        else:
            # For wire compatibility with 3.7 and lower
            header = struct.pack("!i", n)
            if n > 16384:
                # The payload is large so Nagle's algorithm won't be triggered
                # and we'd better avoid the cost of concatenation.
                self._send(header)
                self._send(buf)
            else:
                # Issue #20540: concatenate before sending, to avoid delays due
                # to Nagle's algorithm on a TCP socket.
                # Also note we want to avoid sending a 0-length buffer separately,
                # to avoid "broken pipe" errors if the other end closed the pipe.
                self._send(header + buf)

    # Borrowed from the python3.8 superclass -- handles > 2GB pickles.
    def _recv_bytes(self, maxsize=None):
        buf = self._recv(4)
        size, = struct.unpack("!i", buf.getvalue())
        if size == -1:
            buf = self._recv(8)
            size, = struct.unpack("!Q", buf.getvalue())
        if maxsize is not None and size > maxsize:
            return None
        return self._recv(size)

## This is essential! -- register a special pickler for our
## Connection subclass.  Without this, the file descriptors don't
## get set correctly across the spawn call.
from multiprocessing import reduction
from multiprocessing.connection import reduce_connection
reduction.register(TimingConnection, reduce_connection)

def TimingPipe(track_input, track_output):
    '''
    Creates a pipe composed of Connection or TimingConnection objects,
    depending on what we want to record.
    '''
    fd1, fd2 = os.pipe()
    r = TimingConnection(fd1, writable=False)
    w = TimingConnection(fd2, readable=False)
    return r,w

def _sum_object_stats(O1, O2):
    ''' used to sum the statistics from two objects that may have
    a stats() function that returns a dict.'''
    S1 = S2 = {}
    if hasattr(O1, 'stats'):
        S1 = O1.stats()
    if hasattr(O2, 'stats'):
        S2 = O2.stats()
    # merge/sum
    for k,v in S2.items():
        if not k in S1:
            S1[k] = v
        else:
            S1[k] = S1[k] + S2[k]
    return S1

class TimingSimpleQueue(SimpleQueue):
    '''
    A *SimpleQueue* subclass that uses a *TimingPipe* object to keep
    stats on how much pickling objects costs.
    '''

    # new method
    def stats(self):
        '''
        Returns stats on the objects sent through this queue.
        '''
        return _sum_object_stats(self._reader, self._writer)

    def __init__(self, track_input, track_output, ctx):
        self._reader, self._writer = TimingPipe(track_input, track_output)
        self._rlock = ctx.Lock()
        self._poll = self._reader.poll
        if sys.platform == 'win32':
            self._wlock = None
        else:
            self._wlock = ctx.Lock()

class TimingPool(Pool):
    '''
    A python multiprocessing Pool subclass that keeps track of the
    resources used by workers, and tracks the expense of pickling
    objects.
    '''
    # New functions added:
    def get_worker_cpu(self):
        return self.beancounter.get_cpu()
    def get_worker_wall(self):
        return self.beancounter.get_wall()
    def get_pickle_traffic(self):
        return _sum_object_stats(self._inqueue, self._outqueue)
    def get_pickle_traffic_string(self):
        S = self.get_pickle_traffic()
        return (('  pickled %i objs, %g MB, using %g s CPU\n' +
                 'unpickled %i objs, %g MB, using %g s CPU') %
                 tuple(S.get(k,0) for k in [
                     'pickle_objs', 'pickle_megabytes', 'pickle_cputime',
                     'unpickle_objs', 'unpickle_megabytes', 'unpickle_cputime']))

    def __init__(self, *args,
                 track_send_pickles=True,
                 track_recv_pickles=True,
                 **kwargs):
        self.track_send_pickles = track_send_pickles
        self.track_recv_pickles = track_recv_pickles
        super().__init__(*args, **kwargs)
        self.beancounter = BeanCounter()

    def _setup_queues(self):
        if self.track_send_pickles:
            self._inqueue = TimingSimpleQueue(False, True, self._ctx)
        else:
            self._inqueue = self._ctx.SimpleQueue()
        if self.track_recv_pickles:
            self._outqueue = TimingSimpleQueue(True, False, self._ctx)
        else:
            self._outqueue = self._ctx.SimpleQueue()
        self._quick_get = self._quick_get_wrapper
        self._real_quick_get = self._outqueue._reader.recv
        self._quick_put = self._quick_put_wrapper
        self._real_quick_put = self._inqueue._writer.send

    def _quick_get_wrapper(self):
        # Peel off the timing results
        obj = self._real_quick_get()
        if obj is None:
            return obj
        job, i, (success, res) = obj
        if success:
            res,dt = res
            self.beancounter.add_time(dt)
        return job, i, (success, res)

    def _quick_put_wrapper(self, task):
        # Wrap tasks with timing results
        if task is not None:
            job, i, func, args, kwds = task
            func = time_func(func)
            task = job, i, func, args, kwds
        return self._real_quick_put(task)

class time_func(object):
    '''A wrapper that records the CPU time used by a call, and returns that
    along with the result.'''
    def __init__(self, func):
        self.func = func
    def __call__(self, *args, **kwargs):
        t1 = CpuMeas()
        R = self.func(*args, **kwargs)
        t2 = CpuMeas()
        dt = (t2.cpu_seconds_since(t1), t2.wall_seconds_since(t1))
        return R,dt

######################################################
# These classes handle tracking time & the Measurement framework.

class BeanCounter(object):
    '''
    A class to keep track of the CPU and Wall time used by workers.
    '''
    def __init__(self):
        self.cpu = 0.
        self.wall = 0.
        self.lock = multiprocessing.Lock()

    ### LOCKING
    def add_time(self, dt):
        self.lock.acquire()
        try:
            (cpu, wall) = dt
            self.cpu += cpu
            self.wall += wall
        finally:
            self.lock.release()
    def get_cpu(self):
        self.lock.acquire()
        try:
            return self.cpu
        finally:
            self.lock.release()
    def get_wall(self):
        self.lock.acquire()
        try:
            return self.wall
        finally:
            self.lock.release()
    def __str__(self):
        return 'CPU time: %.3fs s, Wall time: %.3fs' % (self.get_cpu(), self.get_wall())

class TimingPoolMeas(object):
    '''
    An astrometry.util.ttime Measurement object to measure the resources used
    by workers, and by pickling objects.
    '''
    def __init__(self, pool, pickleTraffic=True):
        self.pool = pool
        self.nproc = pool._processes
        self.pickleTraffic = pickleTraffic
    def __call__(self):
        return TimingPoolTimestamp(self.pool, self.pickleTraffic, self.nproc)

class TimingPoolTimestamp(object):
    '''
    The current resources used by a pool of workers, for
    astrometry.util.ttime
    '''
    def __init__(self, pool, pickleTraffic, nproc):
        self.pool = pool
        self.nproc = nproc
        self.t = self.now(pickleTraffic)
        self.cpu = CpuMeas()
    def format_diff(self, other):
        t1 = self.t
        t0 = other.t
        wall = self.cpu.wall_seconds_since(other.cpu)
        main_cpu = self.cpu.cpu_seconds_since(other.cpu)
        worker_cpu = t1['worker_cpu'] - t0['worker_cpu']
        worker_wall = t1['worker_wall'] - t0['worker_wall']
        use = (main_cpu + worker_cpu) / wall
        s = ('%.3f s worker CPU, %.3f s worker Wall, Wall: %.3f s, Cores in use: %.2f, Total efficiency (on %i cores): %.1f %%' %
             (worker_cpu, worker_wall, wall, use, self.nproc, 100.*use / float(self.nproc)))
        if 'pickle_objs' in self.t:
            s += (', pickled %i/%i objs, %.1f/%.1f MB' %
                  tuple(t1[k] - t0[k] for k in [
                        'pickle_objs', 'unpickle_objs',
                        'pickle_megabytes', 'unpickle_megabytes']))
        return s

    def now(self, pickleTraffic):
        if pickleTraffic:
            stats = self.pool.get_pickle_traffic()
        else:
            stats = dict()
        stats.update(worker_cpu = self.pool.get_worker_cpu(),
                     worker_wall = self.pool.get_worker_wall())
        return stats

######################################################



def test_func(N):
    import numpy as np
    r = []
    for i in range(N):
        arr = np.random.normal(size=1000000)
        r.append(arr)
    s = 0.
    for arr in r:
        s = s + arr**2
    return s

def test_func_2(arr):
    import numpy as np
    return np.sum(arr**2)

def test_func_3(arr):
    return arr[::2]**2

def test():
    #import logging
    #multiprocessing.log_to_stderr().setLevel(logging.DEBUG)

    # per multiprocess.Pool documentation (as of Python 3.8 at least), can't just
    # let a Pool fall out of scope; must close or you risk the process hanging.
    # This definitely happens on Linux!
    with TimingPool(4) as pool:
        R = pool.map(test_func, [10, 20, 30])
        print('worker cpu:', pool.get_worker_cpu())
        print('worker wall:', pool.get_worker_wall())
        print('pickles:', pool.get_pickle_traffic_string())

    import numpy as np
    print('Creating second pool')
    with TimingPool(4) as pool:
        print('Using second pool')
        R = pool.map(test_func_2, [np.random.normal(size=1000000) for x in range(5)])
        print('Got result from second pool')
        print('worker cpu:', pool.get_worker_cpu())
        print('worker wall:', pool.get_worker_wall())
        print('pickles:', pool.get_pickle_traffic_string())

    from astrometry.util.ttime import Time
    with TimingPool(4, track_send_pickles=False, track_recv_pickles=False) as pool:
        m = TimingPoolMeas(pool, pickleTraffic=False)
        Time.add_measurement(m)
        t0 = Time()
        R = pool.map(test_func, [20, 20, 20])
        print(Time()-t0)
        Time.remove_measurement(m)

    with TimingPool(4) as pool:
        Time.add_measurement(TimingPoolMeas(pool, pickleTraffic=True))
        t0 = Time()
        R = pool.map(test_func_3, [np.random.normal(size=1000000) for x in range(5)])
        print(Time()-t0)


def test_func_4(arr):
    arr = arr**2
    return arr

def test_jumbo():
    # Test jumbo (> 2 GB) args/results
    import numpy as np
    from astrometry.util.ttime import Time
    with TimingPool(2) as pool:
        Time.add_measurement(TimingPoolMeas(pool, pickleTraffic=True))
        t0 = Time()
        R = pool.map(test_func_4, [np.ones(int(2.1 * 1024 * 1024 * 1024 / 8))])
        print('Jumbo:', np.sum(R))
        print(Time()-t0)

def test_input_generator(n):
    for i in range(n):
        import numpy as np
        x = np.random.random((1000,1000))
        print('Yielding input', i)
        yield (i,x)

def test_sleep(x):
    import time
    time.sleep(3.)
    return x

def test_queue():
    # Generate a bunch of tasks with 1MB pickles...
    in_iter = test_input_generator(100)
    with TimingPool(4) as pool:
        out_iter = pool.imap_unordered(test_sleep, in_iter)
        while True:
            try:
                r = next(out_iter)
                i,x = r
                print('Got result', i)
            except StopIteration:
                print('StopIteration')
                break

if __name__ == '__main__':
    #test_jumbo()
    #test()
    test_queue()
    sys.exit()
